import logging
import os

import fire
import yaml

from .composer import PdfFileComposer
from .utils import get_output_filename

log = logging.getLogger()


def merge_documents(config_filename: str, output_filename: (str, None) = None, overwrite: bool = False,
                    log_level: str = 'INFO') -> None:
    """
    Merges PDF documents using a YAML configuration file.

    :param config_filename: Configuration file path.
    :param output_filename: Optional: Output file name and path. If not specified, uses the name and path of the
     config file with the extension changed to ``.pdf``.
    :param overwrite: By default, an existing output file is not overwritten. Change to ``True`` for overwriting any
     existing files.
    :param log_level: Detail level of output messages. Default is ``INFO``.
    """
    logging.basicConfig(format='%(levelname)s:%(message)s', level=log_level)
    log.info("Reading config from %s", config_filename)
    with open(config_filename, 'rb') as cfg_file:
        config = yaml.safe_load(cfg_file)
    if not output_filename:
        output_filename = get_output_filename(config_filename)
    if os.path.isfile(output_filename) and not overwrite:
        raise FileExistsError("The output file is already present.", output_filename)
    cp = PdfFileComposer(config['contents'], config['metadata'], config['paths'])
    cp.write(output_filename)


def run():
    fire.Fire(merge_documents)


if __name__ == '__main__':
    run()



from collections import namedtuple
import logging
import os
import re

from PyPDF4 import PdfFileReader
from PyPDF4.utils import PdfReadError

PdfInfo = namedtuple('PdfInfo', ['num_pages', 'pdf_version'])

PDF_VERSION_PATTERN = re.compile(r'\d\.\d')
PDF_HEADER_PATTERN = re.compile(r'%PDF-(\d\.\d)')

META_ATTRIBUTES = [
    'author',
    'creator',
    'keywords',
    'producer',
    'subject',
    'title',
]
META_ATTRIBUTE_KEYS = {
    kw: '/{0}'.format(kw.capitalize())
    for kw in META_ATTRIBUTES
}

log = logging.getLogger()


def expand_path(filename: str) -> str:
    """
    Expands variables (user and environment) in a file name.

    :param filename: File name, possibly containing variables.
    :return: File name with variables expanded.
    """
    return os.path.expandvars(os.path.expanduser(filename))


def get_meta_dict(cfg_meta: dict) -> dict:
    """
    Generates a dictionary of metadata that can be added to a PDF file. Invalid keys are skipped with a warning.

    >>> get_meta_dict({'author': 'TheAuthor', 'subject': 'TheSubject', 'title': 'TheTitle', 'invalid': 'Invalid'})
    {'/Author': 'TheAuthor', '/Subject': 'TheSubject', '/Title': 'TheTitle'}

    :param cfg_meta: Input keys and values from config.
    :return: Dictionary containing valid entries for PDF meta data.
    """
    meta_dict = {}
    for k, v in cfg_meta.items():
        meta_key = META_ATTRIBUTE_KEYS.get(k)
        if meta_key:
            meta_dict[meta_key] = v
        else:
            log.warning("Unrecognized metadata key '%s'", meta_key)
    return meta_dict


def get_output_filename(cfg_filename: str) -> str:
    """
    Derives a PDF output file name from the input configuration file name.

    >>> get_output_filename('sample.yaml')
    'sample.pdf'
    >>> get_output_filename('path/to/sample.yaml')
    'path/to/sample.pdf'

    :param cfg_filename: Name and path of the configuration file.
    :return: Same file name and path, with the extension changed to ``.pdf``.
    """
    cfg_base = os.path.splitext(cfg_filename)[0]
    return '{0}.pdf'.format(cfg_base)


def get_pdf_info(input_pdf) -> PdfInfo:
    """
    Checks the version header and retrieves some metadata from the input PDF stream.

    :param input_pdf: File-like object with PDF data.
    :return: A PdfInfo tuple with the number of pages and the PDF version.
    :raises PyPDF4.utils.PdfReadError: If the header of the PDF seems to be invalid.
    """
    raw_header = input_pdf.read(8)
    try:
        header_str = raw_header.decode('utf-8')
    except UnicodeDecodeError:
        raise PdfReadError("PDF header not found")
    header_match = PDF_HEADER_PATTERN.fullmatch(header_str)
    if not header_match:
        raise PdfReadError("PDF header does not seem to be valid")
    pdf_version = header_match.group(1)
    reader = PdfFileReader(input_pdf)
    log.debug("PDF version %s with %d pages", pdf_version, reader.numPages)
    return PdfInfo(reader.numPages, pdf_version)


def get_config_pdf_version(config_version: str, max_input_version: str) -> str:
    """
    From the PDF version as set in the configuration and the maximum version of all input files, checks for
    the best PDF output version. Logs a warning, if the version set in the configuration is lower than any of the
    input files.

    >>> get_config_pdf_version('auto', '1.6')
    '1.6'
    >>> get_config_pdf_version('1.3', '1.5')
    '1.3'
    >>> get_config_pdf_version('1.x', '1.5')
    Traceback (most recent call last):
      ...
    ValueError: ('Invalid PDF version in configuration', '1.x')

    :param config_version: Version string from the configuration. Set to ``auto`` to just use ``max_input_version``.
    :param max_input_version: Maximum version from all input files.
    :return: ``config_version``, unless set to ``auto``, then ``max_input_version``. However, the automatic version
     setting will never be lower than ``1.3``.
    :raises ValueError: If the configuration-set version is an invalid pattern.
    """
    if config_version == 'auto':
        return max(max_input_version, '1.3')
    if not PDF_VERSION_PATTERN.fullmatch(config_version):
        raise ValueError("Invalid PDF version in configuration", config_version)
    if max_input_version > config_version:
        log.warning("PDF version specified in config (%s) is lower than at least one of the input documents (%s). "
                    "The resulting PDF may not be displayed correctly in all viewers.",
                    config_version, max_input_version)
    return config_version



import logging
import os
from collections import namedtuple

from PyPDF4 import PdfFileMerger

from .utils import expand_path, get_pdf_info, get_meta_dict, get_config_pdf_version

LevelInfo = namedtuple('LevelInfo', ['num_pages', 'max_pdf_version'])

log = logging.getLogger()


class PdfFileComposer:
    """
    Wraps a PyPDF.PdfFileMerger for simplifying the merge process of multiple PDF files into a single PDF,
    defined by a configuration.

    :param contents: Content dictionary, optional with nested bookmark structure.
    :param metadata: Output PDF meta data.
    :param search_paths: List of search paths to look up files without any path specification.
    """
    def __init__(self, contents: dict, metadata: dict, search_paths: list):
        self._contents = contents
        if metadata:
            self._output_version = metadata.pop('version', 'auto')
            self._metadata = metadata
        else:
            self._output_version = 'auto'
            self._metadata = None
        self._merger = PdfFileMerger()
        self._resolve_paths = list(map(expand_path, search_paths or ()))

    def _resolve_file(self, filename: str) -> str:
        if os.path.sep in filename:
            return expand_path(filename)
        elif self._resolve_paths:
            log.debug("Looking up file %s", filename)
            # First try in the current directory.
            if os.path.isfile(filename):
                return filename
            for path in self._resolve_paths:
                try_filename = os.path.join(path, filename)
                if os.path.isfile(try_filename):
                    return try_filename
        return filename

    def _set_version(self, max_input_version: str):
        pdf_version = get_config_pdf_version(self._output_version, max_input_version)
        log.debug("Using PDF version %s", pdf_version)
        self._merger.output._header = '%PDF-{0}'.format(pdf_version).encode('utf-8')

    def _set_metadata(self):
        if not self._metadata:
            log.info("No metadata specified; using defaults.")
            return
        meta_dict = get_meta_dict(self._metadata)
        if meta_dict:
            self._merger.addMetadata(meta_dict)

    def _append_content(self, contents: dict, level_base_page: int, parent_bookmark) -> LevelInfo:
        merger = self._merger
        level_pages = 0
        max_pdf_version = ''
        for item in contents:
            bookmark_title = item.get('bookmark')
            if bookmark_title:
                log.info("Inserting bookmark '%s' on page %d", bookmark_title, level_base_page + level_pages)
                this_bookmark = merger.addBookmark(bookmark_title, level_base_page + level_pages,
                                                   parent=parent_bookmark)
            else:
                this_bookmark = parent_bookmark
            document_filename = item.get('document')
            if document_filename:
                resolved_name = self._resolve_file(document_filename)
                log.info("Adding file %s", resolved_name)
                with open(resolved_name, 'rb') as input_pdf:
                    pdf_info = get_pdf_info(input_pdf)
                    max_pdf_version = max(max_pdf_version, pdf_info.pdf_version)
                    level_pages += pdf_info.num_pages
                    merger.append(input_pdf)
                    log.info("%d pages appended", pdf_info.num_pages)
            nested_contents = item.get('contents')
            if nested_contents:
                nested_info = self._append_content(nested_contents, level_base_page + level_pages, this_bookmark)
                max_pdf_version = max(max_pdf_version, nested_info.max_pdf_version)
                level_pages += nested_info.num_pages
        return LevelInfo(level_pages, max_pdf_version)

    def write(self, output_filename: str):
        """
        Merges the configured files into one PDF and writes the result into the specified path.

        :param output_filename: Output PDF name.
        """
        self._set_metadata()
        try:
            res = self._append_content(self._contents, 0, None)
            log.info("Total pages %d - Max PDF version: %s", res.num_pages, res.max_pdf_version)
            if res.num_pages == 0:
                log.warning("No pages have been added.")
            self._set_version(res.max_pdf_version)
            log.info("Writing to file %s", output_filename)
            with open(output_filename, 'wb') as out_file:
                self._merger.write(out_file)
        finally:
            self._merger.close()



import os
from distutils.spawn import find_executable

from setuptools import setup, find_packages

from pdfconfig import __version__


def include_readme():
    try:
        import pandoc
    except ImportError:
        return ''
    pandoc.core.PANDOC_PATH = find_executable('pandoc')
    readme_file = os.path.join(os.path.dirname(__file__), 'README.md')
    doc = pandoc.Document()
    with open(readme_file, 'r') as rf:
        doc.markdown = rf.read()
        return doc.rst


setup(
    name='pdf-config',
    version=__version__,
    packages=find_packages(),
    install_requires=['fire', 'PyPDF4', 'PyYAML'],
    url='https://github.com/merll/pdf-config',
    license='MIT',
    author='Matthias Erll',
    author_email='matthias@erll.de',
    description='Merges PDF files into one using a YAML configuration file.',
    long_description=include_readme(),
    platforms=['OS Independent'],
    keywords=['pdf', 'yaml'],
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    entry_points={
        "console_scripts": [
            "pdfconfig = pdfconfig.main:run",
        ],
    }
)
